!
! Module: l_source_cutouts
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: l_source_cutouts.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module l_source_cutouts

  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use errorHandling

  use coordinate
  use psf_ims
  use wcs_aux
  use constants
  use ellipse

  implicit none 

  private 

  public initializeSrcDataT
  public getSourceCutouts

  interface initializeSrcDataT
    module procedure initializeSrcDataTScalar
    module procedure initializeSrcDataTVector
  end interface

  interface getSourceCutouts
    module procedure getSourceCutoutsFewArgs, getSourceCutoutsManyArgs
  end interface

  type, public :: SrcDataT
    integer(int16) ::&
      idNum
    real(single) ::&
      flux,&		! These three should generally be in compatible
      bkgFluxPerPixel,&	! units. Thus, if flux is cts/s, bkgFlux should
!      bkgFluxPerArcsec,&! be cts/s/pixel.
      xPixel,&
      yPixel,&
      xDet,&
      yDet
    character(10) ::&
      fluxUnit	! eg 'counts', 'rate', 'cgs'
    type(EllipseT) ::&	! Defined in ssclib/ellipse
      region
    real(double) ::&
      raDeg,&
      decDeg,&
      EeV,&
      thetaArcsec,&
      phi,&
      xTan,&
      yTan
    logical(bool) ::&
      isValid
  end type SrcDataT

contains 

  !---------------------------------------------------------------------
  subroutine initializeSrcDataTScalar(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(SrcDataT), intent(out) :: srcData

    type(SrcDataT) :: srcDataVector(1)

    !-------------------------------------------------------------------
    ! Start

    call initializeSrcDataTVector(srcDataVector)
    srcData = srcDataVector(1)

  end subroutine initializeSrcDataTScalar

  !---------------------------------------------------------------------
  subroutine initializeSrcDataTVector(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(SrcDataT), intent(out) :: srcData(:)

    !-------------------------------------------------------------------
    ! Start

    srcData%flux = 0.0
    srcData%bkgFluxPerPixel = 0.0
    srcData%xPixel = 0.0
    srcData%yPixel = 0.0
    srcData%xDet = 0.0
    srcData%yDet = 0.0
    srcData%fluxUnit = 'none'
    call initializeEllipseT(srcData%region)
    srcData%raDeg = 0.0d0
    srcData%decDeg = 0.0d0
    srcData%EeV = 0.0d0
    srcData%thetaArcsec = 0.0d0
    srcData%phi = 0.0d0
    srcData%xTan = 0.0d0
    srcData%yTan = 0.0d0
    srcData%isValid = .false.

  end subroutine initializeSrcDataTVector

  !---------------------------------------------------------------------
  subroutine getSourceCutoutsFewArgs(imgWcs, srcWcs, scAttitude&
  , expStartTimeStamp, fracBackground, srcData)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up with the correct instrument beforehand.

    implicit none

    type(WcsT),                   intent(in)    :: imgWcs,&
                                                   srcWcs
    real(double),                 intent(in)    :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in)    :: scAttitude
    real(single),                 intent(in)    :: fracBackground
    type(SrcDataT),               intent(inout) :: srcData(:)

    integer :: numSources, srcNum
    logical(bool) :: mask(size(srcData))
    real(single) ::&
      bkgFluxesPerSquareArcsec(size(srcData)),&
      integral,&
      ellipseLongAxisArcsec(size(srcData)),&
      ellipseShrtAxisArcsec(size(srcData)),&
      ellipseLongAxisTan(size(srcData)),&
      ellipseShrtAxisTan(size(srcData)),&
      ellipseRotAngle(size(srcData)),&
      ellipseXampTan(size(srcData)), ellipseYampTan(size(srcData)),&
      peakFluxPerSquareArcsec(size(srcData))
    real(single) ::&
      averageBkgFluxPerPixel, srcPixelAreaArcsec
    real(double) ::&
      ctrFovXtan, ctrFovYtan,&
pixelSizeXWorld, pixelSizeYWorld

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcData)

    !...................................................................
    ! Do some checks:

    if (all(srcData%flux <= 0.0)) then
      call warning('allSrcNonPositive', 'None of the sources have rate &
      &values greater than zero.')
  return
    end if

    if (all(srcData%flux <= 0.0 .or. srcData%bkgFluxPerPixel <= 0.0)) then
      call warning('allSrcBkgNonPositive', 'None of the positive-rate &
      &sources have a positive-valued local background.')
  return
    end if

    if (any(srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel <= 0.0)) then
      call warning('nonPositiveSrcBkg', 'Some of the background &
      &values of the source list are less than or equal to zero. These &
      &have been set to the average of the positive values.')

      mask = (srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel > 0.0)
      averageBkgFluxPerPixel = sum(srcData%bkgFluxPerPixel, mask)&
        / real(count(mask))

      where(srcData%flux > 0.0 .and. srcData%bkgFluxPerPixel <= 0.0)&
        srcData%bkgFluxPerPixel = averageBkgFluxPerPixel
    end if

    !...................................................................
    ! Correct bkg flux for pixel size:

!    srcPixelAreaArcsec = abs(srcWcs%X%pixelSizeWorld* srcWcs%Y%pixelSizeWorld)&
!      * 3600.0 * 3600.0
srcPixelAreaArcsec = wcsPixelAreaWorld(srcWcs) * 3600.0 * 3600.0 ! in ssclib/wcs_aux

!**** note this depends on the worldUnits being both 'deg'. Should check this.

    bkgFluxesPerSquareArcsec = srcData%bkgFluxPerPixel / srcPixelAreaArcsec

    !...................................................................
    ! Get src fov coords:

    call raDecToInst(srcData%raDeg, srcData%decDeg, scAttitude&
      , expStartTimeStamp, srcData%xDet, srcData%yDet, srcData%thetaArcsec&
      , srcData%phi)
    ! in ssclib/coordinate.

    !...................................................................
    ! Get peak fluxes:

    do srcNum = 1, numSources
      integral = getPsfIntegral(srcData(srcNum)%thetaArcsec&
        , srcData(srcNum)%EeV)

      peakFluxPerSquareArcsec(srcNum) = srcData(srcNum)%flux / integral
    end do

    !...................................................................
    ! Get ellipse axes:

    call getPsfEllipseAxes(srcData%thetaArcsec, srcData%EeV&
      , peakFluxPerSquareArcsec&
      , bkgFluxesPerSquareArcsec, fracBackground, ellipseLongAxisArcsec&
      , ellipseShrtAxisArcsec, srcData%isValid)

!**** check and report about invalid sources?

    !...................................................................
    ! Convert ra and dec to image tangent coords:

    call raDecToTan(imgWcs%X%refWorld, imgWcs%Y%refWorld, srcData%raDeg&
      , srcData%decDeg, srcData%xTan, srcData%yTan) ! in ssclib/coordinate.

    call instToTan(ctrFovXtan, ctrFovYtan, imgWcs%X%refWorld, imgWcs%Y%refWorld&
      , scAttitude, expStartTimeStamp, thetaArcsec=0.0d0, phi=0.0d0)
    ! in ssclib/coordinate.

    ellipseShrtAxisTan = ellipseShrtAxisArcsec / 3600.0 * PiOn180
    ellipseLongAxisTan = ellipseLongAxisArcsec / 3600.0 * PiOn180
    ellipseRotAngle = atan2(srcData%yTan - ctrFovYtan&
      , srcData%xTan - ctrFovXtan)

    !...................................................................
    ! Convert ellipses from axes and rotAngleDeg to amps and phase:

    do srcNum = 1, numSources
      call ellipseAngleToPhase(ellipseShrtAxisTan(srcNum)&
        , ellipseLongAxisTan(srcNum), ellipseRotAngle(srcNum)&
        , ellipseXampTan(srcNum), ellipseYampTan(srcNum)&
        , srcData(srcNum)%region%phase) ! in ssclib/ellipse
    end do

    !...................................................................
    ! Change image tangents to pixel coords, also scale ellipse amps to
    ! pixel coords:

    call tanToPixels(imgWcs, srcData%xTan, srcData%yTan, srcData%xPixel&
      , srcData%yPixel) ! in ssclib/wcs_aux
!    srcData%region%xAmp = ellipseXampTan / PiOn180&
!      / abs(imgWcs%X%pixelSizeWorld)
!    srcData%region%yAmp = ellipseYampTan / PiOn180&
!      / abs(imgWcs%Y%pixelSizeWorld)
!!*** probably _shouldn't_ have the abs()s here. They also should be got rid of in the tan<->pixels converters in wcs_aux.

pixelSizeXWorld = sqrt(imgWcs%x%matrixOnAxis * imgWcs%x%matrixOnAxis&
 + imgWcs%x%matrixOffAxis * imgWcs%x%matrixOffAxis)
pixelSizeYWorld = sqrt(imgWcs%y%matrixOnAxis * imgWcs%y%matrixOnAxis&
 + imgWcs%y%matrixOffAxis * imgWcs%y%matrixOffAxis)
!*** this is only valid if there is no skew.

srcData%region%xAmp = ellipseXampTan&
  / pixelSizeXWorld / PiOn180
srcData%region%yAmp = ellipseYampTan&
  / pixelSizeYWorld / PiOn180
!**** There are several things wrong with this. Easiest to fix or at least trap is that it assumes worldUnits are both 'deg' and that ctype = 'RA---TAN/DEC--TAN' etc. Harder to trap is that the above assumes there is no skew between pixel and world coords. Finally (altho it's a related issue, if I found a way to handle skew it would also handle this) the parity of the transform gets lost in the way that pixelSizeXWorld are calculated. pixelSizeXWorld and pixelSizeYWorld end up as always positive.

    srcData%region%unit = 'pixels'

    !...................................................................
    ! Convert ellipses from amps and phase to axes and rotAngleDeg:

    do srcNum = 1, numSources
      call ellipsePhaseToAngle(&
        srcData(srcNum)%region%xAmp,&
        srcData(srcNum)%region%yAmp,&
        srcData(srcNum)%region%phase,&
        srcData(srcNum)%region%shortSemiAxis,&
        srcData(srcNum)%region%longSemiAxis,&
        srcData(srcNum)%region%rotatedAngle&
      ) ! in ssclib/ellipse
    end do

  end subroutine getSourceCutoutsFewArgs

  !---------------------------------------------------------------------
  subroutine getSourceCutoutsManyArgs(imgWcs, srcWcs, scAttitude&
  , expStartTimeStamp, fracBackground, srcEeV, srcRaDeg, srcDecDeg, srcRate&
  , srcBkgRatePixels, xPixels, yPixels, ellipseXampPixels, ellipseYampPixels&
  , ellipsePhase, srcIsValid)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up with the correct instrument beforehand.

    implicit none

    type(WcsT),                   intent(in) :: imgWcs,&
                                                srcWcs
    real(double),                 intent(in) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    real(single),                 intent(in) :: fracBackground
    real(double),                 intent(in) :: srcEeV(:),&
                                                srcRaDeg(size(srcEeV)),&
                                                srcDecDeg(size(srcEeV))
    real(single),                 intent(in) :: srcRate(size(srcEeV)),&
                                                srcBkgRatePixels(size(srcEeV))
    real(single),  intent(out) :: xPixels(size(srcEeV)),&
                                  yPixels(size(srcEeV)),&
                                  ellipseXampPixels(size(srcEeV)),&
                                  ellipseYampPixels(size(srcEeV)),&
                                  ellipsePhase(size(srcEeV))
    logical(bool), intent(out) :: srcIsValid(size(srcEeV))

    type(SrcDataT) :: srcData(size(srcEeV))

    !-------------------------------------------------------------------
    ! Start

    srcData%EeV             = srcEeV
    srcData%raDeg           = srcRaDeg
    srcData%decDeg          = srcDecDeg
    srcData%flux            = srcRate
    srcData%bkgFluxPerPixel = srcBkgRatePixels

    call getSourceCutoutsFewArgs(imgWcs, srcWcs, scAttitude&
      , expStartTimeStamp, fracBackground, srcData)

    xPixels           = srcData%xPixel
    yPixels           = srcData%yPixel
    ellipseXampPixels = srcData%region%xAmp
    ellipseYampPixels = srcData%region%yAmp
    ellipsePhase      = srcData%region%phase
    srcIsValid        = srcData%isValid

  end subroutine getSourceCutoutsManyArgs

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_source_cutouts
